function varargout = checkcomments(varargin)
% {RESULTS} = CHECKCOMMENTS({OPTIONS})
%   Compute percentage of lines/characters that are comments within all the m-files in a folder.
%   Print the results in a formatted table.
%
%   This tool counts both the number of lines which contain comments, as well as the number
%   of characters utilized for said comments.  Counting lines may be seen as a description 
%   of the comment frequency.  Consider that this method favors the consistent use of brief 
%   trailing comments.  On the other hand, counting characters gives a description of commentary 
%   volume.  This measure favors full-line comments and descriptive synopses.
%
%   For function files, the size of the help synopsis is also given in absolute and relative terms.
%
%  OPTIONS includes the following key-value pairs:
%    'path' optionally allows the user to explicitly specify a base directory.
%        If unspecified, a dialog is opened to allow the interactive selection of one.
%    'sort' optionally specifies how the output table is sorted (default 'raw')
%        'raw' sorts by the path structure
%        'file' sorts by filename (case-insensitive)
%        'clines' sorts by the number of comment lines
%        'tlines' sorts by the number of total lines
%        'plines' sorts by comment line percentage
%        'cchars' sorts by the number of comment characters
%        'tchars' sorts by the number of total characters
%        'pchars' sorts by comment character percentage
%        'slines' sorts by synopsis size (lines)
%        'schars' sorts by synopsis size (characters)
%    'direction' optionally specifies the sort direction. May be 'ascend' (default) or 'descend'
%        This option has no effect when 'sort' is set to 'raw'.
%    'mincommentary' optionally specifies the minimum allowable commentary percentage (default 10)
%        This parameter can be specified as a scalar or as a vector [%lines %chars]
%        Files which have less commentary will be remarked in a second list for emphasis.
%    'countblank' key specifies that blank lines should contribute to total line count.
%        So as to avoid favoring poorly-spaced code, blank lines are ignored by default.
%        This option also forces the inclusion of whitespace characters in the character counts. 
%    'recursive' key specifies that the search behavior should include all subdirectories.
%    'nosummary' key suppresses printing of overall summary messages.
%
%   RESULTS is a cell array containing the information for all the represented files.
%     This array contains a cellchar vector of filenames and a 12-column numeric 
%     array representing respective fields in the main output table.
%
%  EXAMPLES:
%   Interactively select a directory, collect the results in an output array
%    R = checkcomments();
%
%   Look in PWD, sort by percent with most-commented files on the bottom
%    checkcomments('path','.','sort','percent','dir','ascend')
%
%  This was based on comment_counter.m by Image Analyst, though not much remains.  
%  Support file checkcommProcessor.m heavily relies on work provided by Rik.
% 
% See also: checkmimtdep

% TODO:
% look into trailing linefeed issue with POF causing excess linecount
% pruning might not be a good workaround

% internal configuration
filepattern = '*.m'; % the file pattern used when searching in the specified directories
sortmodestrings = {'raw','file','clines','tlines','plines','slines','cchars','tchars','pchars','schars'}; % used in validating inputs
sortdirstrings = {'ascend','descend'}; % used in validating inputs

% defaults for user inputs
minAllowedPct = [10 10]; % [lines chars]
sortmode = 'raw';
sortdir = 'descend';
recursivesearch = false;
ignoreblanks = true;
printsummary = true;
topLevelFolder = pwd;
explicitpathspec = false;

% parse inputs and override defaults as requested
parseinputs()



% Initialization steps:
clc;    % Clear the command window.

% Ask user to confirm/change the starting folder if none is explicitly given.
if ~explicitpathspec
	topLevelFolder = uigetdir(topLevelFolder,'Pick a starting folder');
	if topLevelFolder == 0
		return;
	end
end
fprintf('The top level folder is "%s".\n\n', topLevelFolder);

% The original path collection method only works in R2016b and later.
% This legacy method should work in older versions (R2015b and R2009b tested)
if recursivesearch
	dirstosearch = regexp(genpath(topLevelFolder),'[^:]*','match').';
else
	dirstosearch = {topLevelFolder};
end
filenames = {};
filefolders = {};
for dk = 1:numel(dirstosearch)
	S = dir(fullfile(dirstosearch{dk},filepattern));
	filenames = [filenames; {S.name}.']; %#ok<*AGROW>

	filefolders = [filefolders; repmat(dirstosearch(dk),numel(S),1)];
end
allFileInfo.folder = filefolders;
allFileInfo.name = filenames;

% Get cell arrays of path and filename strings
listOfFolderNames = unique(allFileInfo.folder);
listOfFileNames = allFileInfo.name;
numfolders = length(listOfFolderNames);
numfiles = length(listOfFileNames);

% barf a message if there are no matches
if numfiles == 0
	if recursivesearch
		fprintf('No matching files found in top-level directory or any of its subdirectories.\n\n')
	else
		fprintf('No matching files found in top-level directory.\n\n')
	end
	return;
end
% Now we have a list of all files, matching the pattern, in the top level folder and its subfolders.



% preallocate
totalcounts = zeros(numfiles,2);
commentcounts = zeros(numfiles,2);
synopsiscounts = zeros(numfiles,2);

% Process all files in those folders.
for kf = 1 : numfiles %#ok<*FXUP>
	% construct path info for this file
	thisFolder = allFileInfo.folder{kf};
	thisBaseFileName = allFileInfo.name{kf};
	fullFileName = fullfile(thisFolder, thisBaseFileName);
		
	% count commentary in this file
	% count arrays are 2 column [lines chars]
	[totalcounts(kf,:), commentcounts(kf,:) synopsiscounts(kf,:)] = checkcommProcessor(fullFileName,ignoreblanks);	
end


percentage = 100*commentcounts./totalcounts;
synpcttot = 100*synopsiscounts./totalcounts;
synpctcom = 100*synopsiscounts./commentcounts;

sumtotalcounts = sum(totalcounts,1);
sumcommentcounts = sum(commentcounts,1);

% sort and print info for each file
maxtlcdigits = max(totalcounts,[],1);
maxtlcdigits = max(floor(log10(abs(maxtlcdigits)))+1,1);
maxfnchars = max(cellfun(@numel,listOfFileNames));
sortfileinfo()
printfileinfo();
printlowcommentlist()

% print overall info for all files
if printsummary
	fprintf('\nDone processing all folders!\nFound %d files in %d folders.\n\n', numfiles, numfolders);

	fprintf('Overall, %d out of %d lines (%.2f%%) had comments.\n', sumcommentcounts(1), sumtotalcounts(1), mean(percentage(:,1)));
	fprintf('Overall, %d out of %d characters (%.2f%%) represented comments.\n\n', sumcommentcounts(2), sumtotalcounts(2), mean(percentage(:,2)));

	fprintf('The average file contained %.0f lines, %.0f of which contained comments.\n', ...
			sumtotalcounts(1)/numfiles, sumcommentcounts(1)/numfiles);
	fprintf('The average file contained %.0f characters, %.0f of which represented comments.\n', ...
			sumtotalcounts(2)/numfiles, sumcommentcounts(2)/numfiles);
		
	fprintf('The average function file synopsis contained %.0f lines, or %.0f characters', ...
			mean(synopsiscounts(~isnan(synopsiscounts(:,1)),:),1)) % 'omitnan' causes version dependency
end
fprintf('\n')

% prepare output if requested
if nargout>0
	% filenames [commentlines totallines %lines commentchars totalchars %chars synlines syn%totl syn%coml synchars syn%totc syn%comc]
	varargout{1} = {listOfFileNames.', [commentcounts(:,1) totalcounts(:,1) percentage(:,1) ...
										commentcounts(:,2) totalcounts(:,2) percentage(:,2) ...
										synopsiscounts(:,1) synpcttot(:,1) synpctcom(:,1) ...
										synopsiscounts(:,2) synpcttot(:,2) synpctcom(:,2)]};
end


%% NESTED FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function sortfileinfo()
	% sort the info table
	switch sortmode
		case 'raw' 
			return; % no sorting
		case 'file'
			% do case-insensitive sort
			firstletters = char(listOfFileNames');
			firstletters = lower(firstletters(:,1));
			[~,idx] = sort(firstletters,sortdir);
		case 'clines'
			[~,idx] = sort(commentcounts(:,1),sortdir);
		case 'tlines'
			[~,idx] = sort(totalcounts(:,1),sortdir);
		case 'plines'
			[~,idx] = sort(percentage(:,1),sortdir);
		case 'cchars'
			[~,idx] = sort(commentcounts(:,2),sortdir);
		case 'tchars'
			[~,idx] = sort(totalcounts(:,2),sortdir);
		case 'pchars'
			[~,idx] = sort(percentage(:,2),sortdir);
		case 'slines'
			[~,idx] = sort(synopsiscounts(:,1),sortdir);
		case 'schars'
			[~,idx] = sort(synopsiscounts(:,2),sortdir);
		otherwise
			% this shouldn't be able to happen since keys are prematched
			error('kablooey')
	end
	
	percentage = percentage(idx,:);
	totalcounts = totalcounts(idx,:);
	commentcounts = commentcounts(idx,:);
	synopsiscounts = synopsiscounts(idx,:);
	synpcttot = synpcttot(idx,:);
	synpctcom = synpctcom(idx,:);
	listOfFileNames = listOfFileNames(idx);
end



function printfileinfo()
	% print info about each file in a formatted table	
	pctfw = 5; % fieldwidth for percent values
	pctdp = 1; % decimal places for percent values
	pctspec = sprintf('%%%d.%df%%%%',pctfw,pctdp);
	
	colwidths = [maxfnchars ...
				2*maxtlcdigits + 7 + pctfw ...
				2*maxtlcdigits(1) + 29 + pctfw ...
				2*maxtlcdigits(2) + 29 + pctfw];
	headparts = {'FILENAME','COMMENT LINES','COMMENT CHARACTERS','SYNOPSIS LINES','SYNOPSIS CHARACTERS'};
	hpsizes = cellfun(@numel,headparts);
	padsz = colwidths - hpsizes;
	
	colpad = '    ';	
	hpad1l = repmat(' ',1,floor(padsz(1)/2)); % pad on LHS of fn
	hpad1r = repmat(' ',1,ceil(padsz(1)/2)); % pad on RHS of fn
	hpad2l = repmat(' ',1,floor(padsz(2)/2)); % pad on LHS of LC
	hpad2r = repmat(' ',1,ceil(padsz(2)/2)); % pad on RHS of LC
	hpad3l = repmat(' ',1,floor(padsz(3)/2)); % pad on LHS of CC
	hpad3r = repmat(' ',1,ceil(padsz(3)/2)); % pad on RHS of CC
	hpad4l = repmat(' ',1,floor(padsz(4)/2)); % pad on LHS of SLC
	hpad4r = repmat(' ',1,ceil(padsz(4)/2)); % pad on RHS of SLC
	hpad5l = repmat(' ',1,floor(padsz(5)/2)); % pad on LHS of SCC
	
	naffstr = 'not a function file';
	npadsz = colwidths(4:5) - numel(naffstr);
	npad1l = repmat(' ',1,floor(npadsz(1)/2)); % pad on LHS of naffstr
	npad1r = repmat(' ',1,ceil(npadsz(1)/2)); % pad on RHS of naffstr
	npad2l = repmat(' ',1,floor(npadsz(2)/2)); % pad on LHS of naffstr
	
	fprintf('%s%s%s%s %s%s%s%s %s%s%s%s %s%s%s%s %s%s\n\n',hpad1l,headparts{1},hpad1r,colpad,hpad2l,headparts{2}, ...
		    hpad2r,colpad,hpad3l,headparts{3},hpad3r,colpad,hpad4l,headparts{4},hpad4r,colpad,hpad5l,headparts{5})
	
	for kk = 1:numfiles
		tlcdigits = max(floor(log10(abs(totalcounts(kk,1))))+1,1); % total line count digits
		tccdigits = max(floor(log10(abs(totalcounts(kk,2))))+1,1); % total char count digits
		
		clcfieldw = num2str(maxtlcdigits(1)); % comment line count fieldwidth
		cccfieldw = num2str(maxtlcdigits(2)); % comment char count fieldwidth
		
		% variable-width padding blocks
		fnpadding = repmat(' ',1,maxfnchars-numel(listOfFileNames{kk})); % trails filename
		tlcpadding = repmat(' ',1,maxtlcdigits(1)-tlcdigits); % trails total line count
		tccpadding = repmat(' ',1,maxtlcdigits(2)-tccdigits); % trails total char count

		% one statement per column block
		fprintf('%s%s%s ',listOfFileNames{kk}, fnpadding, colpad)
		fprintf(['%' clcfieldw 'd / %d%s (' pctspec ')' colpad ' '],commentcounts(kk,1), totalcounts(kk,1), tlcpadding, percentage(kk,1))
		fprintf(['%' cccfieldw 'd / %d%s (' pctspec ')' colpad ' '],commentcounts(kk,2), totalcounts(kk,2), tccpadding, percentage(kk,2))
		
		% synopsis columns
		if ~isnan(synopsiscounts(kk,1))
			fprintf(['%' clcfieldw 'd (' pctspec ' of total  ' pctspec ' of comments)' colpad ' '],synopsiscounts(kk,1), synpcttot(kk,1), synpctcom(kk,1))
			fprintf(['%' cccfieldw 'd (' pctspec ' of total  ' pctspec ' of comments)\n'],synopsiscounts(kk,2), synpcttot(kk,2), synpctcom(kk,2));
		else
			fprintf('%s%s%s%s %s%s\n',npad1l,naffstr,npad1r,colpad,npad2l,naffstr)
		end
	end
	fprintf('\n')
	fprintf('%s%s%s%s %s%s%s%s %s%s%s%s %s%s%s%s %s%s\n\n',hpad1l,headparts{1},hpad1r,colpad,hpad2l,headparts{2}, ...
		    hpad2r,colpad,hpad3l,headparts{3},hpad3r,colpad,hpad4l,headparts{4},hpad4r,colpad,hpad5l,headparts{5})
end



function printlowcommentlist()
	% alert user to files with marginal commentary measure
	% this is more efficiently done in the report dump instead of pestering the user with dialogs
	% it's not as emphatic though
	idx = find(percentage(:,1)<minAllowedPct(1));
	if ~isempty(idx)
		fprintf('\nThe following files have less than %d%% of lines commented (low comment frequency)\n',round(minAllowedPct(1)))
		
		for kk = 1:numel(idx)
			tlcdigits = max(floor(log10(abs(totalcounts(idx(kk),1))))+1,1); % total line count digits
			clcfieldw = num2str(maxtlcdigits(1)); % comment line count fieldwidth

			% variable-width padding blocks
			fnpadding = repmat(' ',1,maxfnchars-numel(listOfFileNames{idx(kk)})); % trails filename
			tlcpadding = repmat(' ',1,maxtlcdigits(1)-tlcdigits); % trails total line count

			fprintf(['%s%s %' clcfieldw 'd / %d%s lines (%5.2f%%)\n'], ...
				listOfFileNames{idx(kk)}, fnpadding, commentcounts(idx(kk),1), totalcounts(idx(kk),1), tlcpadding, percentage(idx(kk),1));
		end
	end
	
	idx = find(percentage(:,2)<minAllowedPct(2));
	if ~isempty(idx)
		fprintf('\nThe following files devote less than %d%% of characters to comments (low comment volume)\n',round(minAllowedPct(2)))
		
		for kk = 1:numel(idx)
			tccdigits = max(floor(log10(abs(totalcounts(idx(kk),2))))+1,1); % total char count digits
			cccfieldw = num2str(maxtlcdigits(2)); % comment char count fieldwidth

			% variable-width padding blocks
			fnpadding = repmat(' ',1,maxfnchars-numel(listOfFileNames{idx(kk)})); % trails filename
			tccpadding = repmat(' ',1,maxtlcdigits(2)-tccdigits); % trails total char count

			fprintf(['%s%s %' cccfieldw 'd / %d%s characters (%5.2f%%)\n'], ...
				listOfFileNames{idx(kk)}, fnpadding, commentcounts(idx(kk),2), totalcounts(idx(kk),2), tccpadding, percentage(idx(kk),2));
		end
	end
end



function parseinputs()
	% parse inputs
	% this could be better, but eh. 
	% i know a lot of validation tools exist, but most of those are new enough
	% that they won't work in the other versions i expect to use anyway.
	% ... assuming i fix the R2016b+ dependency
	if numel(varargin)>0
		ka = 1
		while ka<=numel(varargin)
			thisarg = varargin{ka};
			if ~ischar(thisarg)
				error('CHECKCOMMENTARY: expected argument %d to be a key name',ka)
			end
			switch lower(thisarg)
				case 'path'
					nextarg = varargin{ka+1};
					if ~ischar(nextarg)
						error('CHECKCOMMENTARY: expected value for PATH to be text')
					end
					if ~exist(nextarg,'dir')
						error('CHECKCOMMENTARY: PATH does not appear to be a directory')
					end
					topLevelFolder = nextarg;
					explicitpathspec = true;
					ka = ka+2;
					
				case {'sort','sortmode'}
					nextarg = varargin{ka+1};
					if ~ischar(nextarg) || ~strismember(lower(nextarg),sortmodestrings)
						error('CHECKCOMMENTARY: unknown value for SORTMODE')
					end
					sortmode = lower(nextarg);
					ka = ka+2;
					
				case {'dir','direction'}
					nextarg = varargin{ka+1};
					if ~ischar(nextarg) || ~strismember(lower(nextarg),sortdirstrings)
						error('CHECKCOMMENTARY: unknown value for DIRECTION')
					end
					sortdir = lower(nextarg);
					ka = ka+2;
										
				case 'mincommentary'
					nextarg = varargin{ka+1};
					if ~isnumeric(nextarg)
						error('CHECKCOMMENTARY: expected MINCOMMENTARY value to be numeric')
					end
					
					if isscalar(nextarg)
						minAllowedPct = [1 1]*nextarg;
					else
						minAllowedPct = nextarg;
					end
					ka = ka+2;
									
				case 'recursive'
					recursivesearch = true;
					ka = ka+1;
					
				case 'countblank'
					ignoreblanks = false;
					ka = ka+1;
					
				case 'nosummary'
					printsummary = false;
					ka = ka+1;
					
				otherwise
					error('CHECKCOMMENTARY: unknown key name %s',thisarg)
			end
		end
	end
end


end % END MAIN SCOPE





















